#print "FOO!\n";
package Nexus;
require Exporter;
require DynaLoader;

$version = 1.06;

use Carp;

@ISA = qw(Exporter DynaLoader);

require Nexus::RRT;
require Nexus::RemoteReference;
require Nexus::DeepPacker;
require Nexus::RPCStatus;

#my(%HandlerTable) = ();
%HandlerTable = ();
@PendingRPCList = ();

sub boot
{
    &Nexus::RRT::boot();
    &Nexus::RemoteReference::boot();
    if (defined(&main::boot))
    {
	&main::boot();
    }
}

sub initialize
{
    my(@refs);

    @refs = Nexus::init();

    if (@refs == 0)
    {
	Nexus::start_nonblocking();
    }
    else
    {
	Nexus::start();
    }

    return @refs;
}

sub remote_reference
{
    my($ref) = @_;
    my($rref, $rrt);

    if (!ref($ref))
    {
	croak "Nexus::remote_reference(): Argument must be a reference";
    }

    $rrt = new Nexus::RRT($ref);
    $rref = new Nexus::RemoteReference($rrt->gp(), ref($ref));

    return $rref;
}

sub local
{
    my($rref) = @_;

    if (ref($rref) eq "Nexus::RemoteReference")
    {
	my($gp) = $rref->gp();
	if ($gp->is_local())
	{
	    my($rrt);
	    $rrt = $gp->dereference();
	    return $rrt->ref();
	}
	else
	{
	    return undef;
	}
    }
    elsif (ref($rref) eq "Nexus::GlobalPointer")
    {
	if ($rref->is_local())
	{
	    return $rref->dereference();
	}
	else
	{
	    return undef;
	}
    }
}

sub rpc
{
    my($rref, $name, @args) = @_;
    my($status, $rreftype);

    $rreftype = ref($rref);
    if ($rreftype ne "Nexus::GlobalPointer" &&
	$rreftype ne "Nexus::RemoteReference")
    {
	croak "Nexus::rpc_start(): received non-reference in first argument";
    }

    #
    # Keep a ref to the arg list around, so that
    # any temporary remote refs that were created don't go away
    # between the start and wait.
    #
    $status = new Nexus::RPCStatus($rref, $name, \@args);
#    print "rpc_start creates $status\n";

    my $buf = Nexus::init_rsr($rref, "Nexus::rpc_handler");

    if ($debug > 2)
    {
	print "Before pack:\n";
	Nexus::RRT::dump_rrts();
	Nexus::sv_dump($status);
    }
    
    Nexus::DeepPacker::pack($buf, $name, $status, \@args);

    if ($debug > 2)
    {
	print "After pack:\n";
	Nexus::RRT::dump_rrts();
	Nexus::sv_dump($status);
    }

    $buf->send_rsr();

    my($ret);

    #
    # IVAN'S BUG
    # 
    # Poll once here, in case an earlier poll found the
    # handler invocation we're looking for.
    #
    print "wait $status: finished=", $status->finished(), "\n" if $debug;
    Nexus::poll(0);
    while (!$status->finished())
    {
	print "Nexus::rpc_wait($status): finished=", $status->finished(), " polling\n" if $debug > 1;
	Nexus::poll(1);
    }
    my $rv = $status->return_value();
    return $rv;
#    return $status->return_value();
}

sub rpc_old
{
    my($rref, $name, @args) = @_;
    my($ret);

    {
	my($status);

	$status = &rpc_start($rref, $name, @args);
	print "Nexus::rpc(): status=$status\n" if $debug;
	$ret = &rpc_wait($status);
    }

    print "rpc returning\n";
    return $ret;
}

sub rpc_async
{
    my($rref, $name, @args) = @_;
    my($buf);
    
    my $rreftype = ref($rref);
    if ($rreftype ne "Nexus::GlobalPointer" &&
	$rreftype ne "Nexus::RemoteReference")
    {
	croak "Nexus::rpc_async(): received non-reference in first argument";
    }
	
    $buf = Nexus::init_rsr($rref, $name);
    
    Nexus::DeepPacker::pack($buf, @args);
    return $buf->send_rsr();
}

sub rpc_start
{
    my($rref, $name, @args) = @_;

    my $rreftype = ref($rref);
    if ($rreftype ne "Nexus::GlobalPointer" &&
	$rreftype ne "Nexus::RemoteReference")
    {
	croak "Nexus::rpc_start(): received non-reference in first argument";
    }

    #
    # Keep a ref to the arg list around, so that
    # any temporary remote refs that were created don't go away
    # between the start and wait.
    #
    my $status = new Nexus::RPCStatus($rref, $name, \@args);
    print "rpc_start creates $status\n";

    my $buf = Nexus::init_rsr($rref, "Nexus::rpc_handler");

    Nexus::DeepPacker::pack($buf, $name, $status, \@args);
    $buf->send_rsr();

    return $status;
}

sub rpc_test
{
    my($status) = @_;

    return $status->finished();
}

sub rpc_wait
{
    my($status) = @_;
    my($ret);

    #
    # IVAN'S BUG
    # 
    # Poll once here, in case an earlier poll found the
    # handler invocation we're looking for.
    #
    print "wait $status: finished=", $status->finished(), "\n" if $debug;
    Nexus::poll(0);
    while (!$status->finished())
    {
	print "Nexus::rpc_wait($status): finished=", $status->finished(), " polling\n" if $debug > 1;
	Nexus::poll(1);
    }
    my $rv = $status->return_value();
    return $rv;
#    return $status->return_value();
}

#
# Poll for messages;
#
# If we are blocking,
#    If there are messages pending, don't need to poll nexus with
#    blocking poll.
#
# 
#
sub poll
{
    my($block) = @_;
    my(@handlers, $handler);
    my($nexus_block);

    $nexus_block = $block;
    print "poll($block): pending list: @PendingRPCList\n" if $debug;
    if ($block)
    {
	if (@PendingRPCList == 0)
	{
	    while (@PendingRPCList == 0)
	    {
		print "_poll(1): invoked\n" if $debug > 1;
		&Nexus::_poll(1);
		print "_poll(1): returns\n" if $debug > 1;
	    }
	}
	else
	{
	    #
	    # Poll nexus just to snag messages while we're here
	    #
	    print "_poll(0): invoked\n" if $debug > 1;
	    &Nexus::_poll(0);
	    print "_poll(0): returns\n" if $debug > 1;
	}
    }
    else
    {
	print "_poll(0): invoked\n" if $debug > 1;
	&Nexus::_poll(0);
	print "_poll(0): returns\n" if $debug > 1;
    }

    while (@PendingRPCList)
    {
	$handler = shift(@PendingRPCList);
	print "poll(): got handler $handler @$handler\n" if $debug;
	$handler->handle();
    }
}

sub exit
{
    Nexus::destroy_current_context();
}

#
# register_handler gets a list of (parameter, value) pairs
#


my (%ValidHandlerParameters) = (method_call => 1,
				unpack_buffer => 1);

my (%ValidUnpackValues) = (1 => 1, 0 => 1, old => 1, deep => 1);

sub register_handler
{
    my($handler_name, $func, %params) = @_;
    my($package, $method_call, $auto_unpack);
    my($param, $value);

    while (($param, $value) = each(%params))
    {
	if (!$ValidHandlerParameters{$param})
	{
	    croak "Invalid handler parameter $param\n";
	}
    }

    #
    # Defaults
    #

    $method_call = 0;
    $auto_unpack = 1;

    $method_call = $params{method_call} if defined($params{method_call});
    $auto_unpack = $params{unpack_buffer} if defined($params{unpack_buffer});

    if ($method_call)
    {
	if (!Nexus::isa_string(\$func))
	{
	    croak "Nexus::register_handler: handler function must be a string when method call    is used";
	}
    }
    else
    {
	if (ref($func) !~ /CODE/)
	{
	    croak "Nexus::register_handler: handler function must be a code reference";
	}
    }

    if (!$ValidUnpackValues{$auto_unpack})
    {
	croak "Nexus::register_handler: Invalid unpack type $auto_unpack";
    }
    

    print "Register $handler_name: <$func> <$method_call> <$auto_unpack>\n" if $debug;
    $HandlerTable{$handler_name} = [$func, $method_call, $auto_unpack];
}

sub unregister_handler
{
    my($handler_name) = @_;

    delete $HandlerTable{$handler_name};
}

sub invoke_handler
{
    my($handler_name, $address, $buffer) = @_;
    my($table_entry, $def);

#    $^D = 10 if $handler_name eq 'new_url';
    print "Invoking $handler_name <$address> <$buffer>\n" if $debug;

    $table_entry = $HandlerTable{$handler_name};
    $def = defined($table_entry);
    print "Got table_entry <$table_entry> def=$def\n" if $debug;
    if ($def)
    {
	$default = 0;
    }
    elsif (defined($table_entry = $HandlerTable{$DefaultHandler}))
    {
	$default = 1;
    }
    else
    {
	Carp::croak "Handler $handler_name not defined";
#	die "Handler $handler_name not defined\n";
    }
    {
	my($handler_func, $method_call, $auto_unpack)
	    = @$table_entry;
	my(@args);

	if ($auto_unpack eq 'deep' || $auto_unpack eq 1)
	{
	    my $ref = Nexus::DeepPacker::unpack_buffer($buffer);
	    @args = @$ref;
	}
	elsif ($auto_unpack eq 'old')
	{
	    @args = $buffer->unpack();
	}
	else
	{
	    @args = ($buffer);
	}

	unshift(@args, $handler_name) if $default;

	if ($debug > 1)
	{
	    my($i);
	    for ($i = 0; $i < @args; $i++)
	    {
		print "Arg $i: \n";
		Nexus::sv_dump($args[$i]);
	    }
	}

	print "Calling <$address> <$method_call> <$handler_func>\n" if $debug;
	if ($method_call)
	{
	    $address->$handler_func(@args);
	}
	elsif (1)
	{
	    &$handler_func($address, @args);
	}
	else
	{
	    eval { &$handler_func($address, @args); };

	    if ($@)
	    {
		print "Eval returns <$@>\n";
	    }
	    elsif ($debug)
	    {
		print "Eval returns no error\n";
	    }
	}
    }
}

package Nexus::AsyncHandler;

sub handle
{
    my($self) = @_;
    my($addr, $name, $buf) = @$self;
    print "Async: $addr $name $buf\n" if $debug;

    &Nexus::invoke_handler($name, $addr, $buf);
}

package Nexus::RPCHandler;

use Carp;

sub handle
{
    my($self) = @_;
    my($addr, $buf, $bufdata, $name, $status, $argref, $ret, $rbuf);

    ($addr, $buf) = @$self;

    $bufdata = Nexus::DeepPacker::unpack_buffer($buf);

    ref($bufdata) eq ARRAY or
	die "Nexus::RPCHandler::handle(): Not an ARRAY ref";
    @$bufdata == 3 or
	die "Nexus::RPCHandler::handle(): Wrong array size";
    
    ($name, $status, $argref) = @$bufdata;

    ref($argref) eq ARRAY or
	die "Nexus::RPCHandler::handle(): Not an ARRAY ref";

    print "Nexus::RPCHandler::handle(): name=$name status=$status arg=$argref\n" if $debug;

    $ret = &invoke_rpc($name, $addr, $argref);

    print "Nexus::RPCHandler::handle(): got ret=$ret\n" if $debug;

    $rbuf = Nexus::init_rsr($status, "Nexus::rpc_return_handler");
    Nexus::DeepPacker::pack($rbuf, $ret);
    $rbuf->send_rsr();
    
}

sub invoke_rpc
{
    my($handler_name, $address, $argref) = @_;
    my($table_entry, $def, $ret);

    print "Invoking RPC $handler_name \n" if $debug;

    $table_entry = $Nexus::HandlerTable{$handler_name};
    $def = defined($table_entry);
    print "Got table_entry <$table_entry> def=$def\n" if $debug;
    if ($def)
    {
	$default = 0;
    }
    elsif (defined($table_entry = $Nexus::HandlerTable{$DefaultHandler}))
    {
	$default = 1;
    }
    else
    {
	Carp::croak "Handler $handler_name not defined";
#	die "Handler $handler_name not defined\n";
    }
    {
	my($handler_func, $method_call, $auto_unpack)
	    = @$table_entry;
	my(@args);

	if ($auto_unpack eq 'old')
	{
	    croak "Cannot use old-style handlers with Nexus::rpc()";
	}
	elsif ($auto_unpack ne 'deep' && $auto_unpack ne 1)
	{
	    croak "Must use unpack_buffer=>1 with Nexus::rpc()";
	}

	@args = @$argref;
	unshift(@args, $handler_name) if $default;

	print "Calling <$address> <$method_call> <$handler_func>\n" if $debug;
	if ($method_call)
	{
	    $ret = $address->$handler_func(@args);
	}
	else
	{
	    eval { $ret = &$handler_func($address, @args); };

	    if ($@)
	    {
		print "Eval returns <$@>\n";
	    }
	    elsif ($debug)
	    {
		print "Eval returns no error\n";
	    }
	}
    }
    return $ret;
}

package Nexus::RPCReturnHandler;

sub handle
{
    my($self) = @_;
    my($addr, $buf) = @$self;
    my($ret);

    $ret = Nexus::DeepPacker::unpack_buffer($buf);
    print "Nexus::RPCReturnHandler(): addr=$addr ret=$ret\n" if $debug;
    ref($ret) eq ARRAY or die "Nexus::RPCReturnHandler::handle(): not an ARRAY ref";
    @$ret == 1 or die "Nexus::RPCReturnHandler::handle(): invalid array";
    $addr->set_return_value($ret->[0]);

}

package Nexus::RRTIncr;

sub handle
{
    my($self) = @_;
    my($addr, $value) = @$self;

    print "RRTIncr: self=$self addr=$addr value=$value\n" if $debug;
    if (ref($addr) ne "Nexus::RRT")
    {
	warn "$$: got incr for non-ref $addr\n";
	return;
    }

    $addr->incr($value);
}

sub DESTROY
{
    my($self) = @_;
#    print "RRTIncr::DESTROY $self\n";
}

package Nexus::RRTDecr;

sub handle
{
    my($self) = @_;
    my($addr, $value) = @$self;

    print "RRTDecr: addr=$addr value=$value\n" if $debug;
    if (ref($addr) ne "Nexus::RRT")
    {
	warn "$$: got decr for non-ref $addr\n";
	return;
    }

    $addr->decr($value);
}

sub DESTROY
{
    my($self) = @_;
#    print "RRTDecr::DESTROY $self\n";
}

package Nexus::GlobalPointer;

$OVERLOAD{'""'} = 'as_string';
$OVERLOAD{fallback} = 1;

package Nexus;

bootstrap Nexus;

1;

__END__

