#!/mmlarch/bin/perl

BEGIN { unshift (@INC,qw(./blib .)) }

push(@INC, '/home/mmlab1/Software/nperl/tests');

require Nexus;
require obj;

Nexus::init();
Nexus::start();

Nexus::register_handler("foobar", \&foobar, unpack_buffer => 0);

$var = "hi there";

$gp = new Nexus::GlobalPointer(\$var);

print "init rsr\n";
$buf = Nexus::init_rsr($gp, "foobar");

print "pack $buf\n";

#@data = (1,[2,[a,b,c],3],4,{a=>3,b=>4,c=>[5,6,7]},3.14,7.18);
#@data = ({a => 3, b => 4});

$o = new obj;

@data = ($o);

&dump_nested(\@data);
$p = &pack($buf, @data);

$buf->send_rsr();

Nexus::poll(1);

sub foobar
{
    my($ref, $buf) = @_;
    my($rc);

    print "ref=$ref buf=$buf\n";
    $rc = &::unpack_buffer($buf);
    print "rc=$rc\n";
    &dump_nested($rc);
}

sub dump_nested
{
    my($ref, $indent) = @_;
    my $spaces = ' ' x ($indent * 2);

    if (ref($ref) eq "ARRAY")
    {
	my($l,$r);

	$l = @$ref;
	print "${spaces}Array[$l]:\n";
	for $r (@$ref)
	{
	    &dump_nested($r, $indent + 1);
	}
    }
    elsif (ref($ref) eq "HASH")
    {
	my($k, $v);

	print "${spaces}Hash:\n";
	while (($k, $v) = each(%$ref))
	{
	    print "${spaces}Key=\"$k\"\n";
	    &dump_nested($v, $indent + 1);
	}
    }
    else
    {
	print "${spaces}Other: $ref\n";
    }
}

#
# Packed data structure format
#
# packed_ds: <N items> [items]
#
# item: <type> <n_items> lowlevel_data
#
# type:
#	i	int
#	d	double
#	c	char
#	a	array
#	h	hash
#	r	rref
#
# Notes
#
# The n_chars++ in each of the pack routines is to allocate space for
# the type specifier.
#

sub pack
{
    my($buf, @args) = @_;
    my($arg, $packed_arg);
    my($pack_info, $i);
    my($size, $n_elts);

    $pack_info = {
	items => [],
	n_ints => 0,
	n_chars => 0,
	n_doubles => 0
	n_global_pointers => 0
    };

    &pack_array(\@args, $pack_info);

    &dump_packed($pack_info);

    ($size, $n_elts) = &compute_size($buf, $pack_info);
    print "computed size=$size n_elts=$n_elts\n";
    $buf->set_size($size, $n_elts);
    &pack_buffer($buf, $pack_info);
    
    return $pack_info;
}

sub compute_size
{
    my($buf, $pack_info) = @_;
    my($size, $n_elts);
    my($s, $e);

    $size = $n_elts = 0;

    print "compute: buf=$buf\n";
    ($s, $e) = $buf->sizeof_int($pack_info->{n_ints});
    $size += $s;
    $n_elts += $e;

    ($s, $e) = $buf->sizeof_char($pack_info->{n_chars});
    $size += $s;
    $n_elts += $e;

    ($s, $e) = $buf->sizeof_double($pack_info->{n_doubles});
    $size += $s;
    $n_elts += $e;

    return ($size, $n_elts);
}

sub pack_buffer
{
    my($buf, $pack_info) = @_;
    my($item, $ref, $type, $count);

    for $item (@{$pack_info->{items}})
    {
	($type, $count, $ref) = @$item;

	print "put_string $type\n";
	$buf->put_string($type);

	print "put_int $count\n";
	$buf->put_int($count);

	if ($type eq 'i')
	{
	    $buf->put_int($$ref);
	    print "put_int $$ref\n";
	}
	elsif ($type eq 'd')
	{
	    $buf->put_double($$ref);
	    print "put_double $$ref\n";
	}
	elsif ($type eq 'c')
	{
	    $buf->put_string($$ref);
	    print "put_char $$ref\n";
	}
    }
}

sub unpack_buffer
{
    my($buf, $indent) = @_;
    my($rc, $type, $count, $elt, $key, $val, $i);

    my $spaces = ' ' x ($indent * 2);

    $buf->get_string(\$type, 1);
    print $spaces, "got type $type\n";
    
    $buf->get_int(\$count);
    print $spaces, "got count $count\n";
    
    if ($type eq 'i')
    {
	$count == 1 or die "Nperl can't do arrays of ints yet\n";
	$buf->get_int(\$rc);
	print $spaces, "got int $rc\n";
    }
    elsif ($type eq 'c')
    {
	$buf->get_string(\$rc, $count);
	print $spaces, "got string $rc\n";
    }
    elsif ($type eq 'd')
    {
	$count == 1 or die "Nperl can't do arrays of doubles yet\n";
	$buf->get_double(\$rc);
	print $spaces, "got double $rc\n";
    }
    elsif ($type eq 'h')
    {
	$rc = {};
	for ($i = 0; $i < $count; $i++)
	{
	    $key = &unpack_buffer($buf, $indent + 1);
	    $val = &unpack_buffer($buf, $indent + 1);
	    $rc->{$key} = $val;
	}
    }
    elsif ($type eq 'a')
    {
	$rc = [];
	for ($i = 0; $i < $count; $i++)
	{
	    $elt = &unpack_buffer($buf, $indent + 1);
	    push(@$rc, $elt);
	}
    }
    
    else
    {
	die "Unknown type $type\n";
    }
    return $rc;
}

sub dump_packed
{
    my($pack_info) = @_;
    my($item, $ref, $type, $count, $key);

    for $key (qw(n_ints n_chars n_doubles))
    {
	print "$key = $pack_info->{$key}\n";
    }

    for $item (@{$pack_info->{items}})
    {
	($type, $count, $ref) = @$item;
	print "    $type $count $ref ";
	print $$ref if ref($ref);
	print "\n";
    }
}

sub pack_item
{
    my($item, $pack_info) = @_;
    my($ref);

    $ref = ref($$item);
    if ($ref eq "ARRAY")
    {
	&pack_array($$item, $pack_info);
    }
    elsif ($ref eq "HASH")
    {
	&pack_hash($$item, $pack_info);
    }
    elsif ($ref eq "Nexus::GlobalPointer")
    {
	&pack_global_pointer($$tem, $pack_info);
    }
    elsif ($ref eq "Nexus::RemoteReference")
    {
	&pack_remote_reference($$tem, $pack_info);
    }
    elsif ($ref ne '')
    {
	&pack_ref($$item, $pack_info);
    }
    elsif (Nexus::isa_string($item))
    {
	&pack_string($item, $pack_info);
    }
    elsif (Nexus::isa_int($item))
    {
	&pack_int($item, $pack_info);
    }
    elsif (Nexus::isa_double($item))
    {
	&pack_double($item, $pack_info);
    }
    else
    {
	die "Unknown type for item $item\n";
    }
}

sub pack_global_pointer
{
    my($gp, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['g', 1, $gp]);
    $pack_info->{n_chars}++;
    $pack_info->{n_ints}++;
    $pack_info->{n_global_pointers}++;
}

#
# Remote reference is packed as name & global pointer.
#
sub pack_remote_reference
{
    my($rref, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['r', 1, $rref->gp()]);
    $pack_info->{n_chars}++;
    $pack_info->{n_ints}++;
    $pack_info->{n_global_pointers}++;

    #
    # RRefs send along the type name. We don't use pack_string
    # here.
    #

    my $refname = $rref->type();

    $pack_info->{n_chars} += length($refname);
    $pack_info->{n_ints}++;

    print "pack remote refernece to $refname gp is ", $rref->gp(), "\n";
}

#
# Pack an arbitrary reference.
#
# This is done by creating a remote reference trap for this ref, 
# and sending along the trap's global pointer using the same
# protocol as pack_remote_reference.
#
# This is easy as RRT has the same signature as RemoteReference. :-)
#
sub pack_ref
{
    my($ref, $pack_info) = @_;
    my($rrt);

    $rrt = new Nexus::RRT($ref);

    &pack_remote_reference($rrt, $pack_info);
}

sub pack_int
{
    my($item, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['i', 1, $item]);
    $pack_info->{n_chars}++;

    $pack_info->{n_ints} += 2;
}

sub pack_double
{
    my($item, $pack_info) = @_;

    push(@{$pack_info->{items}}, ['d', 1, $item]);
    $pack_info->{n_chars}++;

    $pack_info->{n_ints}++;
    $pack_info->{n_doubles}++;
}

sub pack_string
{
    my($item, $pack_info) = @_;
    my($len) = length($$item);
    
    push(@{$pack_info->{items}}, ['c', $len, $item]);
    $pack_info->{n_chars}++;

    $pack_info->{n_ints}++;
    $pack_info->{n_chars} += $len;
}

sub pack_array
{
    my($array, $pack_info) = @_;
    my($item, $len, $i);

    $len = $#$array + 1;

    push(@{$pack_info->{items}}, ['a', $len, undef]);
    $pack_info->{n_chars}++;
    $pack_info->{n_ints}++;

    for ($i = 0; $i < $len; $i++)
    {
	&pack_item(\$$array[$i], $pack_info);
    }
}

sub pack_hash
{
    my($hash, $pack_info) = @_;
    my($item, $len, @keys, $key, $elt);

    @keys = keys(%$hash);
    $len = @keys;

    push(@{$pack_info->{items}}, ['h', $len, undef]);
    $pack_info->{n_chars}++;
    $pack_info->{n_ints}++;

    for $key (@keys)
    {
	print "key=$key\n";
	&pack_item(\$key, $pack_info);
	&pack_item(\$hash->{$key}, $pack_info);
    }
}
