#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

#undef PNX_DEBUG 

SV *pnx_rref_new(SV *obj, char *type)
{
    SV *rref;
    SV *cls_sv, *type_sv;
    char *cls;
    STRLEN len;
    dSP;
    /* Need the following for XPUSHp. */
    /* dATARGET; */
    int count;

    cls = "Nexus::RemoteReference";
    len = strlen(cls);

    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(sv_2mortal(newSVpv(cls, len)));
    XPUSHs(obj);
    XPUSHs(sv_2mortal(newSVpv(type, strlen(type))));
    PUTBACK;
    count = perl_call_method("new", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    rref = POPs;
    SvREFCNT_inc(rref);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return rref;
}

SV *pnx_rrt_new(SV *obj)
{
    SV *rrt;
    SV *cls_sv;
    char *cls;
    STRLEN len;
    dSP;
    /* Need the following for XPUSHp. */
    /* dATARGET; */
    int count;

    cls = "Nexus::RRT";
    len = strlen(cls);

    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(sv_2mortal(newSVpv(cls, len)));
    XPUSHs(obj);
    PUTBACK;
    count = perl_call_method("new", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    rrt = POPs;
    SvREFCNT_inc(rrt);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return rrt;
}

SV *pnx_rrt_gp(SV *rrt)
{
    SV *gp;
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("gp", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    gp = POPs;
    SvREFCNT_inc(gp);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return gp;
}

char *pnx_rrt_type(SV *rrt)
{
    char *type;
    static char buf[2048];
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("type", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    type = POPp;
    if (strlen(type) > (sizeof(buf) - 1))
	croak("rrt_type: name %s too long", type);
    strcpy(buf, type);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return buf;
}

SV *pnx_rrt_ref(SV *rrt)
{
    SV *ref;
    
    dSP;
    int count;
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    PUTBACK;
    count = perl_call_method("ref", G_SCALAR | G_EVAL);
    SPAGAIN;
    if (count != 1)
	croak("Big trouble");
    ref = POPs;
    SvREFCNT_inc(ref);
    PUTBACK;
    FREETMPS;
    LEAVE;
    return ref;
}

void pnx_rrt_incr(SV *rrt, int count)
{
    dSP;
    SV *countsv;

    countsv = newSViv(count);
    
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    XPUSHs(sv_2mortal(countsv));
    PUTBACK;
    perl_call_method("incr", G_SCALAR | G_EVAL);
    SPAGAIN;
    PUTBACK;
    FREETMPS;
    LEAVE;
}

void pnx_rrt_decr(SV *rrt, int count)
{
    dSP;
    SV *countsv;

    countsv = newSViv(count);
    ENTER;
    SAVETMPS;
    PUSHMARK(sp);
    XPUSHs(rrt);
    XPUSHs(sv_2mortal(countsv));
    PUTBACK;
    perl_call_method("decr", G_SCALAR | G_EVAL);
    SPAGAIN;
    PUTBACK;
    FREETMPS;
    LEAVE;
}

void pnx_mb_pack_item(SV *bufref, SV **items, int n_items)
{
    pnx_buffer_t *pnx_buffer;
    nexus_buffer_t *buffer;
    int i, gp_i;
    SV *sv, *svr;
    char *typespec, *t;
    int n_ints, n_doubles, n_chars, n_refs;
    char *sval;
    STRLEN len;
    int ilen;
    int ival;
    double dval;
    int size, n_elts;
    char *refname, *objname;
    nexus_global_pointer_t *gps;

    pnx_buffer = _pnx_mb_get_buffer(bufref, "pnx_pack");
    buffer = &pnx_buffer->buffer;

    NexusMalloc(pnx_pack, typespec, char *, sizeof(char) * (n_items + 1));

    n_ints = n_chars = n_doubles = n_refs = 0;

    /* For the typespec */
    n_ints++;
    n_chars += n_items;

    for (i = 0; i < n_items; i++)
    {
	sv = items[i];

	if (!SvOK(sv))
	{
	    typespec[i] = 'u';
	}
	else if (SvPOKp(sv))
	{
	    n_ints++;
	    sval = SvPV(sv, len);
	    n_chars += len;
	    typespec[i] = 's';
	}
	else if (SvIOKp(sv))
	{
	    n_ints++;
	    typespec[i] = 'i';
	}
	else if (SvNOKp(sv))
	{
	    n_doubles++;
	    typespec[i] = 'n';	    
	}
	/*
	 * Argument is a reference.
	 *
	 * If its an object, look to see if its a RemoteReference. If it
	 * is, typespec is 'r' and we pass the global pointer to the RRT
	 * that (returned by sv->gp).
	 *
	 * If it is a non-RR object or is not an object, create an
	 * RRT & pass a global pointer to the RRT. Typespec is 'R' to
	 * denote to the second pass that the RRT must be created. 
	 *
	 * If it is a GlobalPointer object just pass the gp. Typespec is
	 * 'g'. This shouldn't normally happen, but is supported for
	 * low-level nperl hacking.
	 *
	 * If typespec is 'r', we'll pass a type name thru, so bump 
	 * the int & char counters appropriately.
	 *
	 * Note the construction of these objects doesn't happen yet,
	 * as this is just the first pass.
	 *
	 */
	else if (SvROK(sv))
	{
	    svr = SvRV(sv);

	    if (SvOBJECT(svr))
	    {
		refname = HvNAME(SvSTASH(svr));
		
#ifdef PNX_DEBUG
		printf("arg %d is an object of type %s\n", i, refname);
#endif

		if (strcmp(refname, "Nexus::GlobalPointer") == 0)
		{
#ifdef PNX_DEBUG
		    printf("have global pointer\n");
#endif
		    typespec[i] = 'g';
		}
		else if (strcmp(refname, "Nexus::RemoteReference") == 0)
		{
#ifdef PNX_DEBUG
		    printf("have remote ref\n");
#endif
		    typespec[i] = 'r';
		}
		else
		{
#ifdef PNX_DEBUG
		    printf("have other ref\n");
#endif
		    typespec[i] = 'R';
		}
	    }
	    else
	    {
#ifdef PNX_DEBUG
		printf("have nonobject %d\n", i);
#endif
		typespec[i] = 'R';
	    }

	    n_refs++;
	}
	else
	{
	    croak("HELP! Unknown object");
	}
	    
    }
    typespec[n_items] = '\0';
#ifdef PNX_DEBUG
    printf("Typespec: <%s>\n", typespec);
#endif

    NexusMalloc(pnx_pack, gps, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t) * n_refs);
    /*
     * Take a pass through the refs and gps, creating RRTs as necessary
     * and filling in the gps[] array.
     *
     * Also update n_ints and n_chars for the type names to be passed
     * through.
     */

    for (gp_i = 0, i = 0; i < n_items; i++)
    {
	sv = items[i];
	if (typespec[i] == 'R')
	{
	    /*
	     * Create a RRT for this guy, and snag its
	     * gp to pack in the buffer.
	     */
	    SV *rrt, *gpsv;
	    nexus_global_pointer_t *my_gp;

	    rrt = pnx_rrt_new(sv);
	    gpsv = pnx_rrt_gp(rrt);

	    objname = pnx_rrt_type(rrt);

#ifdef PNX_DEBUG
	    printf("rrt objname is %s\n", objname);
#endif
	    n_ints++;
	    n_chars += strlen(objname);
		    
	    my_gp = pnx_gp_get_gp(gpsv, "pnx_pack");
	    gps[gp_i++] = *my_gp;

	    /*
	     * Overwrite the items[i] entry for him, so
	     * we can grab the name later. We don't need
	     * the original sv anymore, and if we did,
	     * we could get it back with rrt_ref().
	     */
	    items[i] = rrt;
	}
	else if (typespec[i] == 'r')
	{
	    SV *gpsv;
	    nexus_global_pointer_t *my_gp;
	    
	    objname = pnx_rrt_type(sv);
#ifdef PNX_DEBUG
	    printf("rref objname=%s\n", objname);
#endif

	    gpsv = pnx_rrt_gp(sv);

	    n_ints++;
	    n_chars += strlen(objname);
		    
	    my_gp = pnx_gp_get_gp(gpsv, "pnx_pack");
	    gps[gp_i++] = *my_gp;
	}
	else if (typespec[i] == 'g')
	{
	    nexus_global_pointer_t *my_gp;
	    my_gp = pnx_gp_get_gp(sv, "pnx_pack");
	    gps[gp_i++] = *my_gp;
	}
    }

    if (n_refs > 0)
	size = nexus_sizeof_global_pointer(buffer, gps, n_refs, &n_elts);
    else
    {
	size = n_elts = 0;
    }

#ifdef PNX_DEBUG
    printf("n_doubles=%d n_ints=%d n_chars=%d n_refs=%d\n",
	   n_doubles, n_ints, n_chars, n_refs);
#endif

    size += nexus_sizeof_int(buffer, n_ints);
    n_elts += n_ints;
    
    size += nexus_sizeof_char(buffer, n_chars);
    n_elts += n_chars;
    
    size += nexus_sizeof_double(buffer, n_doubles);
    n_elts += n_doubles;

    nexus_set_buffer_size(buffer, size, n_elts);

    nexus_put_int(buffer, &n_items, 1);
    nexus_put_char(buffer, typespec, n_items);
    
    for (gp_i = 0, i = 0; i < n_items; i++)
    {
	sv = items[i];

	switch (typespec[i])
	{
	case 'u':
	    break;
	    
	case 's':
	    sval = SvPV(sv, len);
	    ilen = (int) len;
	    nexus_put_int(buffer, &ilen, 1);
	    nexus_put_char(buffer, sval, ilen);
	    break;

	case 'i':
	    ival = SvIV(sv);
	    nexus_put_int(buffer, &ival, 1);
	    break;
	    
	case 'n':
	    dval = SvNV(sv);
	    nexus_put_double(buffer, &dval, 1);
	    break;

	case 'r':
	case 'R':
	    objname = pnx_rrt_type(sv);
	    ilen = strlen(objname);
#ifdef PNX_DEBUG
	    printf("packing spec %c name=<%s> len=%d\n", typespec[i],
		   objname, ilen);
#endif
	    nexus_put_int(buffer, &ilen, 1);
	    nexus_put_char(buffer, objname, ilen);
	    nexus_put_global_pointer(buffer, &gps[gp_i], 1);
	    gp_i++;
	    break;

	case 'g':
	    nexus_put_global_pointer(buffer, &(gps[gp_i]), 1);
	    gp_i++;
	    break;

	default:
	    croak("Invalid typespec %c", typespec[i]);
	    break;
	}
    }
    NexusFree(typespec);
    NexusFree(gps);
}

void pnx_mb_pack(SV *bufref, SV **items, int n_items)
{
    pnx_buffer_t *pnx_buffer;
    nexus_buffer_t *buffer;
    int i, gp_i;
    SV *sv, *svr;
    char *typespec, *t;
    int n_ints, n_doubles, n_chars, n_refs;
    char *sval;
    STRLEN len;
    int ilen;
    int ival;
    double dval;
    int size, n_elts;
    char *refname, *objname;
    nexus_global_pointer_t *gps;

    pnx_buffer = _pnx_mb_get_buffer(bufref, "pnx_pack");
    buffer = &pnx_buffer->buffer;

    NexusMalloc(pnx_pack, typespec, char *, sizeof(char) * (n_items + 1));

    n_ints = n_chars = n_doubles = n_refs = 0;

    /* For the typespec */
    n_ints++;
    n_chars += n_items;

    for (i = 0; i < n_items; i++)
    {
	sv = items[i];

	if (!SvOK(sv))
	{
	    typespec[i] = 'u';
	}
	else if (SvPOKp(sv))
	{
	    n_ints++;
	    sval = SvPV(sv, len);
	    n_chars += len;
	    typespec[i] = 's';
	}
	else if (SvIOKp(sv))
	{
	    n_ints++;
	    typespec[i] = 'i';
	}
	else if (SvNOKp(sv))
	{
	    n_doubles++;
	    typespec[i] = 'n';	    
	}
	/*
	 * Argument is a reference.
	 *
	 * If its an object, look to see if its a RemoteReference. If it
	 * is, typespec is 'r' and we pass the global pointer to the RRT
	 * that (returned by sv->gp).
	 *
	 * If it is a non-RR object or is not an object, create an
	 * RRT & pass a global pointer to the RRT. Typespec is 'R' to
	 * denote to the second pass that the RRT must be created. 
	 *
	 * If it is a GlobalPointer object just pass the gp. Typespec is
	 * 'g'. This shouldn't normally happen, but is supported for
	 * low-level nperl hacking.
	 *
	 * If typespec is 'r', we'll pass a type name thru, so bump 
	 * the int & char counters appropriately.
	 *
	 * Note the construction of these objects doesn't happen yet,
	 * as this is just the first pass.
	 *
	 */
	else if (SvROK(sv))
	{
	    svr = SvRV(sv);

	    if (SvOBJECT(svr))
	    {
		refname = HvNAME(SvSTASH(svr));
		
#ifdef PNX_DEBUG
		printf("arg %d is an object of type %s\n", i, refname);
#endif

		if (strcmp(refname, "Nexus::GlobalPointer") == 0)
		{
#ifdef PNX_DEBUG
		    printf("have global pointer\n");
#endif
		    typespec[i] = 'g';
		}
		else if (strcmp(refname, "Nexus::RemoteReference") == 0)
		{
#ifdef PNX_DEBUG
		    printf("have remote ref\n");
#endif
		    typespec[i] = 'r';
		}
		else
		{
#ifdef PNX_DEBUG
		    printf("have other ref\n");
#endif
		    typespec[i] = 'R';
		}
	    }
	    else
	    {
#ifdef PNX_DEBUG
		printf("have nonobject %d\n", i);
#endif
		typespec[i] = 'R';
	    }

	    n_refs++;
	}
	else
	{
	    croak("HELP! Unknown object");
	}
	    
    }
    typespec[n_items] = '\0';
#ifdef PNX_DEBUG
    printf("Typespec: <%s>\n", typespec);
#endif

    NexusMalloc(pnx_pack, gps, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t) * n_refs);
    /*
     * Take a pass through the refs and gps, creating RRTs as necessary
     * and filling in the gps[] array.
     *
     * Also update n_ints and n_chars for the type names to be passed
     * through.
     */

    for (gp_i = 0, i = 0; i < n_items; i++)
    {
	sv = items[i];
	if (typespec[i] == 'R')
	{
	    /*
	     * Create a RRT for this guy, and snag its
	     * gp to pack in the buffer.
	     */
	    SV *rrt, *gpsv;
	    nexus_global_pointer_t *my_gp;

	    rrt = pnx_rrt_new(sv);
	    gpsv = pnx_rrt_gp(rrt);

	    objname = pnx_rrt_type(rrt);

#ifdef PNX_DEBUG
	    printf("rrt objname is %s\n", objname);
#endif
	    n_ints++;
	    n_chars += strlen(objname);
		    
	    my_gp = pnx_gp_get_gp(gpsv, "pnx_pack");
	    gps[gp_i++] = *my_gp;

	    /*
	     * Overwrite the items[i] entry for him, so
	     * we can grab the name later. We don't need
	     * the original sv anymore, and if we did,
	     * we could get it back with rrt_ref().
	     */
	    items[i] = rrt;
	}
	else if (typespec[i] == 'r')
	{
	    SV *gpsv;
	    nexus_global_pointer_t *my_gp;
	    
	    objname = pnx_rrt_type(sv);
#ifdef PNX_DEBUG
	    printf("rref objname=%s\n", objname);
#endif

	    gpsv = pnx_rrt_gp(sv);

	    n_ints++;
	    n_chars += strlen(objname);
		    
	    my_gp = pnx_gp_get_gp(gpsv, "pnx_pack");
	    gps[gp_i++] = *my_gp;
	}
	else if (typespec[i] == 'g')
	{
	    nexus_global_pointer_t *my_gp;
	    my_gp = pnx_gp_get_gp(sv, "pnx_pack");
	    gps[gp_i++] = *my_gp;
	}
    }

    if (n_refs > 0)
	size = nexus_sizeof_global_pointer(buffer, gps, n_refs, &n_elts);
    else
    {
	size = n_elts = 0;
    }

#ifdef PNX_DEBUG
    printf("n_doubles=%d n_ints=%d n_chars=%d n_refs=%d\n",
	   n_doubles, n_ints, n_chars, n_refs);
#endif

    size += nexus_sizeof_int(buffer, n_ints);
    n_elts += n_ints;
    
    size += nexus_sizeof_char(buffer, n_chars);
    n_elts += n_chars;
    
    size += nexus_sizeof_double(buffer, n_doubles);
    n_elts += n_doubles;

    nexus_set_buffer_size(buffer, size, n_elts);

    nexus_put_int(buffer, &n_items, 1);
    nexus_put_char(buffer, typespec, n_items);
    
    for (gp_i = 0, i = 0; i < n_items; i++)
    {
	sv = items[i];

	switch (typespec[i])
	{
	case 'u':
	    break;
	    
	case 's':
	    sval = SvPV(sv, len);
	    ilen = (int) len;
	    nexus_put_int(buffer, &ilen, 1);
	    nexus_put_char(buffer, sval, ilen);
	    break;

	case 'i':
	    ival = SvIV(sv);
	    nexus_put_int(buffer, &ival, 1);
	    break;
	    
	case 'n':
	    dval = SvNV(sv);
	    nexus_put_double(buffer, &dval, 1);
	    break;

	case 'r':
	case 'R':
	    objname = pnx_rrt_type(sv);
	    ilen = strlen(objname);
#ifdef PNX_DEBUG
	    printf("packing spec %c name=<%s> len=%d\n", typespec[i],
		   objname, ilen);
#endif
	    nexus_put_int(buffer, &ilen, 1);
	    nexus_put_char(buffer, objname, ilen);
	    nexus_put_global_pointer(buffer, &gps[gp_i], 1);
	    gp_i++;
	    break;

	case 'g':
	    nexus_put_global_pointer(buffer, &(gps[gp_i]), 1);
	    gp_i++;
	    break;

	default:
	    croak("Invalid typespec %c", typespec[i]);
	    break;
	}
    }
    NexusFree(typespec);
    NexusFree(gps);
}

void pnx_smb_unpack(SV *bufref, SV ***itemsp, int *n_items_ptr)
{
    pnx_rsr_t *rsr;
    nexus_stashed_buffer_t *buffer;
    int len, n_items;
    char *typespec;
    int i;
    int ival;
    double dval;
    char *sval;
    SV **items;
    nexus_global_pointer_t *gp;
    SV *gp_sv;
    
    rsr = _pnx_smb_get_rsr(bufref, "pnx_unpack");
    buffer = &rsr->u.user.buffer;

    nexus_get_stashed_int(buffer, &n_items, 1);

    *n_items_ptr = n_items;

    NexusMalloc(pnx_pack, items, SV **, sizeof(SV *) * n_items);

    *itemsp = items;
    
    NexusMalloc(pnx_pack, typespec, char *, sizeof(char) * (n_items + 1));

    nexus_get_stashed_char(buffer, typespec, n_items);
    typespec[n_items] = '\0';

/*    printf("got n_items=%d typespec=<%s>\n", n_items, typespec);*/

    for (i = 0; i < n_items; i++)
    {
	switch (typespec[i])
	{
	case 'r':
	case 'R':
	    nexus_get_stashed_int(buffer, &len, 1);
	    NexusMalloc(pnx_pack, sval, char *, sizeof(char) * (len + 1));
	    nexus_get_stashed_char(buffer, sval, len);
	    sval[len] = '\0';
	    
#ifdef PNX_DEBUG
	    printf("got spec %c name %s\n", typespec[i], sval);
#endif

	    NexusMalloc(pnx_unpack, gp, nexus_global_pointer_t *,
			sizeof(nexus_global_pointer_t));
	    nexus_get_stashed_global_pointer(buffer, gp, 1);

	    /*
	     * If the global pointer points to this context,
	     * return the actual reference (from the RRT).
	     * We bump the refcount on the reference itself,
	     * but not on the RRT (since we haven't created a
	     * RemoteReference whose deallocation would decrement
	     * the RRT's refcount.
	     *
	     * Otherwise, create a RemoteReference object
	     * and return that.
	     *
	     */
	    if (nexus_global_pointer_to_current_context(gp))
	    {
		SV *rrt, *ref;
		rrt = nexus_convert_global_pointer_address(gp);

		/*
		 * Here, rrt is the actual object, not a reference to
		 * it. So we can go in and pull out the ref directly
		 * (treating the RRT as the array that it is).
		 */
		
		ref = *(av_fetch((AV *) rrt, (I32) 1, 0));

		items[i] = ref;
		nexus_destroy_global_pointer(gp);
		NexusFree(gp);
	    }
	    else
	    {
		SV *asv;
		gp_sv = sv_setref_iv(newSVsv(&sv_undef),
				     "Nexus::GlobalPointer", (IV) gp);

		gp_sv = sv_2mortal(gp_sv);

		asv = pnx_rref_new(gp_sv, sval);
		items[i] = sv_2mortal(asv);
	    }
	    NexusFree(sval);
	    break;
	    
	case 'g':
	    NexusMalloc(pnx_pack, gp, nexus_global_pointer_t *,
			sizeof(nexus_global_pointer_t));
	    nexus_get_stashed_global_pointer(buffer, gp, 1);
	    items[i] = sv_setref_iv(newSVsv(&sv_undef),
				 "Nexus::GlobalPointer", (IV) gp);
	    break;
	    
	case 'u':
	    items[i] = newSVsv(&sv_undef);
	    break;
	    
	case 's':
	    nexus_get_stashed_int(buffer, &len, 1);
	    NexusMalloc(pnx_pack, sval, char *, sizeof(char) * len);
	    nexus_get_stashed_char(buffer, sval, len);
		   
	    items[i] = newSVpv(sval, len);
	    NexusFree(sval);
	    break;

	case 'i':
	    nexus_get_stashed_int(buffer, &ival, 1);
	    items[i] = newSViv(ival);
	    break;

	case 'n':
	    nexus_get_stashed_double(buffer, &dval, 1);
	    items[i] = newSVnv(dval);
	    break;

	default:
	    croak("Unknown typespec %c", typespec[i]);
	}
    }
    NexusFree(typespec);
}

