#include "EXTERN.h"
#include "perl.h"

#ifndef DBL_DIG
#define DBL_DIG 15
#endif


void pnx_sv_dump(SV* sv)
{
    char tmpbuf[1024];
    char *d = tmpbuf;
    U32 flags;
    U32 type;
    FILE *fp = stdout;

    if (!sv) {
	fprintf(fp, "SV = 0\n");
	return;
    }
    
    flags = SvFLAGS(sv);
    type = SvTYPE(sv);

    sprintf(d, "(0x%lx)\n  REFCNT = %ld\n  FLAGS = (",
	(unsigned long)SvANY(sv), (long)SvREFCNT(sv));
    d += strlen(d);
    if (flags & SVs_PADBUSY)	strcat(d, "PADBUSY,");
    if (flags & SVs_PADTMP)	strcat(d, "PADTMP,");
    if (flags & SVs_PADMY)	strcat(d, "PADMY,");
    if (flags & SVs_TEMP)	strcat(d, "TEMP,");
    if (flags & SVs_OBJECT)	strcat(d, "OBJECT,");
    if (flags & SVs_GMG)	strcat(d, "GMG,");
    if (flags & SVs_SMG)	strcat(d, "SMG,");
    if (flags & SVs_RMG)	strcat(d, "RMG,");
    d += strlen(d);

    if (flags & SVf_IOK)	strcat(d, "IOK,");
    if (flags & SVf_NOK)	strcat(d, "NOK,");
    if (flags & SVf_POK)	strcat(d, "POK,");
    if (flags & SVf_ROK)	strcat(d, "ROK,");
    if (flags & SVf_OOK)	strcat(d, "OOK,");
    if (flags & SVf_FAKE)	strcat(d, "FAKE,");
    if (flags & SVf_READONLY)	strcat(d, "READONLY,");
    d += strlen(d);

    if (flags & SVp_IOK)	strcat(d, "pIOK,");
    if (flags & SVp_NOK)	strcat(d, "pNOK,");
    if (flags & SVp_POK)	strcat(d, "pPOK,");
    if (flags & SVp_SCREAM)	strcat(d, "SCREAM,");
    d += strlen(d);
    if (d[-1] == ',')
	d--;
    *d++ = ')';
    *d = '\0';

    fprintf(fp, "SV = ");
    switch (type) {
    case SVt_NULL:
	fprintf(fp,"NULL%s\n", tmpbuf);
	return;
    case SVt_IV:
	fprintf(fp,"IV%s\n", tmpbuf);
	break;
    case SVt_NV:
	fprintf(fp,"NV%s\n", tmpbuf);
	break;
    case SVt_RV:
	fprintf(fp,"RV%s\n", tmpbuf);
	break;
    case SVt_PV:
	fprintf(fp,"PV%s\n", tmpbuf);
	break;
    case SVt_PVIV:
	fprintf(fp,"PVIV%s\n", tmpbuf);
	break;
    case SVt_PVNV:
	fprintf(fp,"PVNV%s\n", tmpbuf);
	break;
    case SVt_PVBM:
	fprintf(fp,"PVBM%s\n", tmpbuf);
	break;
    case SVt_PVMG:
	fprintf(fp,"PVMG%s\n", tmpbuf);
	break;
    case SVt_PVLV:
	fprintf(fp,"PVLV%s\n", tmpbuf);
	break;
    case SVt_PVAV:
	fprintf(fp,"PVAV%s\n", tmpbuf);
	break;
    case SVt_PVHV:
	fprintf(fp,"PVHV%s\n", tmpbuf);
	break;
    case SVt_PVCV:
	fprintf(fp,"PVCV%s\n", tmpbuf);
	break;
    case SVt_PVGV:
	fprintf(fp,"PVGV%s\n", tmpbuf);
	break;
    case SVt_PVFM:
	fprintf(fp,"PVFM%s\n", tmpbuf);
	break;
    case SVt_PVIO:
	fprintf(fp,"PVIO%s\n", tmpbuf);
	break;
    default:
	fprintf(fp,"UNKNOWN%s\n", tmpbuf);
	return;
    }
    if (type >= SVt_PVIV || type == SVt_IV)
	fprintf(fp, "  IV = %ld\n", (long)SvIVX(sv));
    if (type >= SVt_PVNV || type == SVt_NV)
	fprintf(fp, "  NV = %.*g\n", DBL_DIG, SvNVX(sv));
    if (SvROK(sv)) {
	fprintf(fp, "  RV = 0x%lx\n", (long)SvRV(sv));
	pnx_sv_dump(SvRV(sv));
	return;
    }
    if (type < SVt_PV)
	return;
    if (type <= SVt_PVLV) {
	if (SvPVX(sv))
	    fprintf(fp, "  PV = 0x%lx \"%s\"\n  CUR = %ld\n  LEN = %ld\n",
		(long)SvPVX(sv), SvPVX(sv), (long)SvCUR(sv), (long)SvLEN(sv));
	else
	    fprintf(fp, "  PV = 0\n");
    }
    if (type >= SVt_PVMG) {
	if (SvMAGIC(sv)) {
	    fprintf(fp, "  MAGIC = 0x%lx\n", (long)SvMAGIC(sv));
	}
	if (SvSTASH(sv))
	    fprintf(fp, "  STASH = %s\n", HvNAME(SvSTASH(sv)));
    }
    switch (type) {
    case SVt_PVLV:
	fprintf(fp, "  TYPE = %c\n", LvTYPE(sv));
	fprintf(fp, "  TARGOFF = %ld\n", (long)LvTARGOFF(sv));
	fprintf(fp, "  TARGLEN = %ld\n", (long)LvTARGLEN(sv));
	fprintf(fp, "  TARG = 0x%lx\n", (long)LvTARG(sv));
	pnx_sv_dump(LvTARG(sv));
	break;
    case SVt_PVAV:
	fprintf(fp, "  ARRAY = 0x%lx\n", (long)AvARRAY(sv));
	fprintf(fp, "  ALLOC = 0x%lx\n", (long)AvALLOC(sv));
	fprintf(fp, "  FILL = %ld\n", (long)AvFILL(sv));
	fprintf(fp, "  MAX = %ld\n", (long)AvMAX(sv));
	fprintf(fp, "  ARYLEN = 0x%lx\n", (long)AvARYLEN(sv));
	if (AvREAL(sv))
	    fprintf(fp, "  FLAGS = (REAL)\n");
	else
	    fprintf(fp, "  FLAGS = ()\n");
	break;
    case SVt_PVHV:
	fprintf(fp, "  ARRAY = 0x%lx\n",(long)HvARRAY(sv));
	fprintf(fp, "  KEYS = %ld\n", (long)HvKEYS(sv));
	fprintf(fp, "  FILL = %ld\n", (long)HvFILL(sv));
	fprintf(fp, "  MAX = %ld\n", (long)HvMAX(sv));
	fprintf(fp, "  RITER = %ld\n", (long)HvRITER(sv));
	fprintf(fp, "  EITER = 0x%lx\n",(long) HvEITER(sv));
	if (HvPMROOT(sv))
	    fprintf(fp, "  PMROOT = 0x%lx\n",(long)HvPMROOT(sv));
	if (HvNAME(sv))
	    fprintf(fp, "  NAME = \"%s\"\n", HvNAME(sv));
	break;
    case SVt_PVFM:
    case SVt_PVCV:
	fprintf(fp, "  STASH = 0x%lx\n", (long)CvSTASH(sv));
	fprintf(fp, "  START = 0x%lx\n", (long)CvSTART(sv));
	fprintf(fp, "  ROOT = 0x%lx\n", (long)CvROOT(sv));
	fprintf(fp, "  XSUB = 0x%lx\n", (long)CvXSUB(sv));
	fprintf(fp, "  XSUBANY = %ld\n", (long)CvXSUBANY(sv).any_i32);
	fprintf(fp, "  FILEGV = 0x%lx\n", (long)CvFILEGV(sv));
	fprintf(fp, "  DEPTH = %ld\n", (long)CvDEPTH(sv));
	fprintf(fp, "  PADLIST = 0x%lx\n", (long)CvPADLIST(sv));
	fprintf(fp, "  OUTSIDE = 0x%lx\n", (long)CvOUTSIDE(sv));
	if (type == SVt_PVFM)
	    fprintf(fp, "  LINES = %ld\n", (long)FmLINES(sv));
	break;
    case SVt_PVGV:
	fprintf(fp, "  NAME = %s\n", GvNAME(sv));
	fprintf(fp, "  NAMELEN = %ld\n", (long)GvNAMELEN(sv));
	fprintf(fp, "  STASH = 0x%lx\n", (long)GvSTASH(sv));
	fprintf(fp, "  GP = 0x%lx\n", (long)GvGP(sv));
	fprintf(fp, "    SV = 0x%lx\n", (long)GvSV(sv));
	fprintf(fp, "    REFCNT = %ld\n", (long)GvREFCNT(sv));
	fprintf(fp, "    IO = 0x%lx\n", (long)GvIOp(sv));
	fprintf(fp, "    FORM = 0x%lx\n", (long)GvFORM(sv));
	fprintf(fp, "    AV = 0x%lx\n", (long)GvAV(sv));
	fprintf(fp, "    HV = 0x%lx\n", (long)GvHV(sv));
	fprintf(fp, "    CV = 0x%lx\n", (long)GvCV(sv));
	fprintf(fp, "    CVGEN = 0x%lx\n", (long)GvCVGEN(sv));
	fprintf(fp, "    LASTEXPR = %ld\n", (long)GvLASTEXPR(sv));
	fprintf(fp, "    LINE = %ld\n", (long)GvLINE(sv));
	fprintf(fp, "    FLAGS = 0x%x\n", (int)GvFLAGS(sv));
	fprintf(fp, "    STASH = 0x%lx\n", (long)GvSTASH(sv));
	fprintf(fp, "    EGV = 0x%lx\n", (long)GvEGV(sv));
	break;
    case SVt_PVIO:
	fprintf(fp, "  IFP = 0x%lx\n", (long)IoIFP(sv));
	fprintf(fp, "  OFP = 0x%lx\n", (long)IoOFP(sv));
	fprintf(fp, "  DIRP = 0x%lx\n", (long)IoDIRP(sv));
	fprintf(fp, "  LINES = %ld\n", (long)IoLINES(sv));
	fprintf(fp, "  PAGE = %ld\n", (long)IoPAGE(sv));
	fprintf(fp, "  PAGE_LEN = %ld\n", (long)IoPAGE_LEN(sv));
	fprintf(fp, "  LINES_LEFT = %ld\n", (long)IoLINES_LEFT(sv));
	fprintf(fp, "  TOP_NAME = %s\n", IoTOP_NAME(sv));
	fprintf(fp, "  TOP_GV = 0x%lx\n", (long)IoTOP_GV(sv));
	fprintf(fp, "  FMT_NAME = %s\n", IoFMT_NAME(sv));
	fprintf(fp, "  FMT_GV = 0x%lx\n", (long)IoFMT_GV(sv));
	fprintf(fp, "  BOTTOM_NAME = %s\n", IoBOTTOM_NAME(sv));
	fprintf(fp, "  BOTTOM_GV = 0x%lx\n", (long)IoBOTTOM_GV(sv));
	fprintf(fp, "  SUBPROCESS = %ld\n", (long)IoSUBPROCESS(sv));
	fprintf(fp, "  TYPE = %c\n", IoTYPE(sv));
	fprintf(fp, "  FLAGS = 0x%lx\n", (long)IoFLAGS(sv));
	break;
    }
}
