#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

/*
 * Stuff for select() style notification of pending messages.
 */
static int notify_r_fd, notify_w_fd;
static FILE *notify_fp;
static nexus_mutex_t notify_lock;

static SV *fd_add_notifier = (SV *) NULL;
static SV *fd_remove_notifier = (SV *) NULL;
static SV *fd_add_notifier_arg = (SV *) NULL;
static SV *fd_remove_notifier_arg = (SV *) NULL;

static void fd_add_notifier_func(int);
static void fd_remove_notifier_func(int);

static SV *fd_saved_notifier_fds[1024];

void pnx_fd_boot()
{
    nexus_mutex_init(&notify_lock, (nexus_mutexattr_t *) NULL);

    notify_r_fd = -1;
    notify_w_fd = -1;
    notify_fp = (FILE *) NULL;

    bzero(fd_saved_notifier_fds, sizeof(fd_saved_notifier_fds));
}

FILE *pnx_get_notify_fh()
{
    nexus_mutex_lock(&notify_lock);
    if (notify_fp == (FILE *) NULL)
    {
	int p[2];
	pipe(p);
	notify_r_fd = p[0];
	notify_w_fd = p[1];

	notify_fp = fdopen(notify_r_fd, "r");
    }
    nexus_mutex_unlock(&notify_lock);
    return notify_fp;
}

void pnx_register_fd_notifiers(SV *add_ref, SV *add_arg,
			       SV *remove_ref, SV *remove_arg)
{
    SV *add, *remove;
/*    printf("register fd notifiers\n");*/

    if (SvROK(add_ref))
    {
	add = SvRV(add_ref);
    }
    else
    {
	croak("register_fd_notifiers needs code ref");
    }

    if (SvROK(remove_ref))
    {
	remove = SvRV(remove_ref);
    }
    else
    {
	croak("register_fd_notifiers needs code ref");
    }

    fd_add_notifier = add;
    fd_add_notifier_arg = add_arg;
    fd_remove_notifier = remove;
    fd_remove_notifier_arg = remove_arg;

    SvREFCNT_inc(add_arg);
    SvREFCNT_inc(remove_arg);

    SvREFCNT_inc(add);
    SvREFCNT_inc(remove);

    nexus_register_fd_notifiers(fd_add_notifier_func, fd_remove_notifier_func);
}

static void fd_add_notifier_func(int fd)
{
    dSP;
    GV *gv;
    SV *sv;
    FILE *fp;
    
    
/*    printf("C got add %d\n", fd);*/
/*    pnx_sv_dump(fd_add_notifier_arg);*/

    if ((fp = fdopen(fd, "r")) == (FILE *) NULL)
    {
	croak("fd_add_notifier_func: fdopen failed");
    }
    setbuf(fp, NULL);
    gv = newGVgen("Nexus");
    if (!do_open(gv, "<&", 2, fp))
    {
	croak("fd_add_notifier_func: do_open failed");
    }

    sv = sv_bless(newRV((SV *)gv), gv_stashpv("Nexus", 1));

    SvREFCNT_inc(sv);

    fd_saved_notifier_fds[fd] = sv;

    PUSHMARK(sp);
    XPUSHs(fd_add_notifier_arg);
    XPUSHs(sv_2mortal(sv));
    PUTBACK;
    perl_call_sv(fd_add_notifier, G_EVAL | G_DISCARD);
}

static void fd_remove_notifier_func(int fd)
{
    dSP;
/*    printf("C got remove %d\n", fd);*/
    PUSHMARK(sp);
    XPUSHs(fd_remove_notifier_arg);
    XPUSHs(fd_saved_notifier_fds[fd]);

    PUTBACK;
    perl_call_sv(fd_remove_notifier, G_EVAL | G_DISCARD);

    SvREFCNT_dec(fd_saved_notifier_fds[fd]);
    fd_saved_notifier_fds[fd] = NULL;
}

