#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

void pnx_mb_boot(void)
{
}

void pnx_mb_sizeof_int(SV *self, int n_ints, int *size, int *n_elts)
{
    pnx_buffer_t *buffer;

    buffer = _pnx_mb_get_buffer(self, "pnx_mb_sizeof_int");
    if (!buffer)
    {
	*size = *n_elts = -1;
	return;
    }

    *size = nexus_sizeof_int(&buffer->buffer, n_ints);
    *n_elts = n_ints;
}

void pnx_mb_sizeof_double(SV *self, int n_doubles, int *size, int *n_elts)
{
    pnx_buffer_t *buffer;

    buffer = _pnx_mb_get_buffer(self, "pnx_mb_sizeof_double");
    if (!buffer)
    {
	*size = *n_elts = -1;
	return;
    }

    *size = nexus_sizeof_double(&buffer->buffer, n_doubles);
    *n_elts = n_doubles;
}

void pnx_mb_sizeof_char(SV *self, int n_chars, int *size, int *n_elts)
{
    pnx_buffer_t *buffer;

    buffer = _pnx_mb_get_buffer(self, "pnx_mb_sizeof_char");
    if (!buffer)
    {
	*size = *n_elts = -1;
	return;
    }

    *size = nexus_sizeof_char(&buffer->buffer, n_chars);
    *n_elts = n_chars;
}

void pnx_mb_sizeof_global_pointer(SV *self, SV *gpsv, int *size, int *n_elts)
{
    pnx_buffer_t *buffer;
    nexus_global_pointer_t *gp;

    gp = pnx_gp_get_gp(gpsv, "pnx_mb_sizeof_global_pointer");
    if (!gp)
    {
	*size = *n_elts = -1;
	return;
    }

    buffer = _pnx_mb_get_buffer(self, "pnx_mb_sizeof_char");
    if (!buffer)
    {
	*size = *n_elts = -1;
	return;
    }

    *size = nexus_sizeof_global_pointer(&buffer->buffer, gp, 1, n_elts);
}

void pnx_mb_set_size(SV *self, int size, int n_elts)
{
    pnx_buffer_t *buffer;

    buffer = _pnx_mb_get_buffer(self, "pnx_mb_set_size");
    if (!buffer)
	return;

    nexus_set_buffer_size(&buffer->buffer, size, n_elts);
}

void pnx_mb_put_int(SV *self, int intval)
{
    pnx_buffer_t *buffer;

    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_put_int")))
	return;
    nexus_put_int(&buffer->buffer, &intval, 1);
}

void pnx_mb_put_double(SV *self, double doubleval)
{
    pnx_buffer_t *buffer;

    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_put_double")))
	return;
    nexus_put_double(&buffer->buffer, &doubleval, 1);
}

void pnx_mb_put_string(SV *self, char *string, int len)
{
    pnx_buffer_t *buffer;

    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_put_string")))
	return;
    nexus_put_char(&buffer->buffer, string, len);
}

void pnx_mb_put_global_pointer(SV *self, SV *gpsv)
{
    pnx_buffer_t *buffer;
    nexus_global_pointer_t *gp;
    
    gp = pnx_gp_get_gp(gpsv, "pnx_mb_put_global_pointer");
    if (!gp)
	return;

    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_put_global_pointer")))
	return;
    nexus_put_global_pointer(&buffer->buffer, gp, 1);
}

int pnx_mb_send_rsr(SV *self)
{
    pnx_buffer_t *buffer;

    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_send_rsr")))
	return -1;
    return nexus_send_remote_service_request(&buffer->buffer);
}

void pnx_mb_DESTROY(SV *self)
{
    pnx_buffer_t *buffer;
	
    if (!(buffer = _pnx_mb_get_buffer(self, "pnx_mb_DESTROY")))
        return;
/*    printf("free mb %x\n", buffer);*/
    NexusFree(buffer);
}
	
	

pnx_buffer_t *_pnx_mb_get_buffer(SV *self, char *func)
{
    pnx_buffer_t *buffer;
    SV *bufaddr;
    HV *stash;
    char *package;
    
    if (!SvROK(self))
    {
	croak("Must pass Nexus::MessageBuffer to %s", func);
	return (pnx_buffer_t *) NULL;
    }

    bufaddr = SvRV(self);

    if (!SvOK(bufaddr))
    {
	croak("Must pass Nexus::MessageBuffer to %s", func);
	return (pnx_buffer_t *) NULL;
    }

    stash = SvSTASH(bufaddr);
    if (!stash)
    {
	croak("Must pass Nexus::MessageBuffer to %s", func);
	return (pnx_buffer_t *) NULL;
    }

    package = HvNAME(stash);

    if (strcmp(package, "Nexus::MessageBuffer") != 0)
    {
	croak("Must pass Nexus::MessageBuffer to %s", func);
	return (pnx_buffer_t *) NULL;
    }
    buffer = (pnx_buffer_t *) SvIV(bufaddr);
    return buffer;
}
